# ▄████▄    ██████ ▓█████▄  ▒█████   █     █░███▄    █   ██████  ▄████▄   ▄▄▄       ██▓    ▓█████ 
#▒██▀ ▀█  ▒██    ▒ ▒██▀ ██▌▒██▒  ██▒▓█░ █ ░█░██ ▀█   █ ▒██    ▒ ▒██▀ ▀█  ▒████▄    ▓██▒    ▓█   ▀ 
#▒▓█    ▄ ░ ▓██▄   ░██   █▌▒██░  ██▒▒█░ █ ░█▓██  ▀█ ██▒░ ▓██▄   ▒▓█    ▄ ▒██  ▀█▄  ▒██░    ▒███   
#▒▓▓▄ ▄██▒  ▒   ██▒░▓█▄   ▌▒██   ██░░█░ █ ░█▓██▒  ▐▌██▒  ▒   ██▒▒▓▓▄ ▄██▒░██▄▄▄▄██ ▒██░    ▒▓█  ▄ 
#▒ ▓███▀ ░▒██████▒▒░▒████▓ ░ ████▓▒░░░██▒██▓▒██░   ▓██░▒██████▒▒▒ ▓███▀ ░ ▓█   ▓██▒░██████▒░▒████▒
#░ ░▒ ▒  ░▒ ▒▓▒ ▒ ░ ▒▒▓  ▒ ░ ▒░▒░▒░ ░ ▓░▒ ▒ ░ ▒░   ▒ ▒ ▒ ▒▓▒ ▒ ░░ ░▒ ▒  ░ ▒▒   ▓▒█░░ ▒░▓  ░░░ ▒░ ░
#  ░  ▒   ░ ░▒  ░ ░ ░ ▒  ▒   ░ ▒ ▒░   ▒ ░ ░ ░ ░░   ░ ▒░░ ░▒  ░ ░  ░  ▒     ▒   ▒▒ ░░ ░ ▒  ░ ░ ░  ░
#░        ░  ░  ░   ░ ░  ░ ░ ░ ░ ▒    ░   ░    ░   ░ ░ ░  ░  ░  ░          ░   ▒     ░ ░      ░   
#░ ░            ░     ░        ░ ░      ░            ░       ░  ░ ░            ░  ░    ░  ░   ░  ░
#░                  ░                                           ░                                 
## TODO: Remove once CSDownscale is on CRAN
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Interpolation.R')
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Intbc.R')
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Intlr.R')
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Analogs.R')
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Utils.R')
source("https://earth.bsc.es/gitlab/external/cstools/-/raw/master/R/CST_BiasCorrection.R")


## Entry params data and recipe?
downscale_datasets <- function(recipe, data) {
  # Function that downscale the hindcast using the method stated in the
  # recipe. For the moment, forecast must be null.
  #
  # data: list of s2dv_cube objects containing the hcst, obs and fcst.
  # recipe: object obtained when passing the .yml recipe file to read_yaml()
  
  type <- tolower(recipe$Analysis$Workflow$Downscaling$type)

  if (!is.null(data$fcst)) {
    warning("The downscaling will be only performed to the hindcast data")
    data$fcst <- NULL
  }

  if (type == "none") {
    
    hcst_downscal <- data$hcst
    DOWNSCAL_MSG <- "##### NO DOWNSCALING PERFORMED #####"

  } else {
    # Downscaling function params
    int_methods     <- tolower(recipe$Analysis$Workflow$Downscaling$int_method)
    bc_methods      <- tolower(recipe$Analysis$Workflow$Downscaling$bc_method)
    lr_methods      <- tolower(recipe$Analysis$Workflow$Downscaling$lr_method)
    log_reg_methods <- tolower(recipe$Analysis$Workflow$Downscaling$log_reg_method)
    target_grid     <- tolower(recipe$Analysis$Workflow$Downscaling$target_grid)
    nanalogs        <- as.numeric(recipe$Analysis$Workflow$Downscaling$nanalogs)
    
    if (is.null(recipe$Analysis$ncores)) {
      ncores <- 1
    } else {
      ncores <- recipe$Analysis$ncores
    }
 
    #TO DO: add the parametre loocv where it corresponds
    if (is.null(recipe$Analysis$loocv)) {
      loocv <- TRUE
    } else {
      loocv <- recipe$Analysis$loocv
    }

    DOWNSCAL_TYPES <- c("none", "int", "intbc", "intlr", "analogs", "logres")
    BC_METHODS <- c("simple_bias", "calibration", "quantile_mapping")
    LR_METHODS <- c("basic", "large_scale", "4nn")
    LOG_REG_METHODS <- c("ens_mean", "ens_mean_sd", "sorted_members")

    if (!(type %in% DOWNSCAL_TYPES)) {
      stop("Downscaling type in the recipe is not available. Accepted types ",
           "are 'none', 'int', 'intbc', 'intlr', 'analogs', 'logres'.")
    } 

    if (type == "int") {
      if (is.null(int_methods)) {
        stop("Please provide one or more interpolation methods in the recipe.")
      }

      if (is.null(target_grid)) {
        stop("Please provide the target grid in the recipe.")
      }

      hcst_downscal <- list()
      for (int_method in strsplit(int_methods, ", | |,")[[1]]) {

        hcst_d <- CST_Interpolation(data$hcst, 
                                    points = NULL, 
                                    method_remap = int_method, 
                                    target_grid = target_grid,
                                    lat_dim = "latitude", 
                                    lon_dim = "longitude", 
                                    region = NULL,
                                    method_point_interp = NULL)

        hcst_downscal[[ int_method ]] <- hcst_d
      }  

      DOWNSCAL_MSG <- "##### DOWNSCALING COMPLETE #####"
    } else if (type == "intbc") {
      if (is.null(int_methods) | (length(strsplit(int_methods, ", | |,")[[1]]) != 1)) {
        stop("Please provide one (and only one) interpolation method in the recipe.")
      }

      if (is.null(bc_methods)) {
        stop("Please provide one bias-correction method in the recipe. Accepted ",  
              "methods are 'simple_bias', 'calibration', 'quantile_mapping'.")
      }

      if (is.null(target_grid)) {
        stop("Please provide the target grid in the recipe.")
      }

      hcst_downscal <- list()
      for (bc_method in strsplit(bc_methods, ", | |,")[[1]]) { 

        if (!(bc_method %in% BC_METHODS)) {
          stop(paste0(bc_method, " method in the recipe is not available. Accepted methods ",
               "are 'simple_bias', 'calibration', 'quantile_mapping'."))
        }

        hcst_d <- CST_Intbc(data$hcst, data$obs, 
                            target_grid = target_grid, 
                            bc_method = bc_method, 
                            int_method = int_methods, 
                            points = NULL,
                            method_point_interp = NULL, 
                            lat_dim = "latitude", 
                            lon_dim = "longitude",
                            sdate_dim = "syear", 
                            member_dim = "ensemble", 
                            region = NULL, 
                            ncores = ncores)

        hcst_downscal[[ bc_method ]] <- hcst_d
      }   
 
      DOWNSCAL_MSG <- "##### DOWNSCALING COMPLETE #####"
    } else if (type == "intlr") {

      if (is.null(int_methods) | (length(strsplit(int_methods, ", | |,")[[1]]) != 1)) {
        stop("Please provide one (and only one) interpolation method in the recipe.")
      }

      if (is.null(lr_methods)) {
        stop("Please provide one linear regression method in the recipe. Accepted ", 
              "methods are 'basic', 'large_scale', '4nn'.")
      }

      if (is.null(target_grid)) {
        stop("Please provide the target grid in the recipe.")
      }

      hcst_downscal <- list()
      for (lr_method in strsplit(lr_methods, ", | |,")[[1]]) {

        if (!(lr_method %in% LR_METHODS)) {
          stop(paste0(lr_method, " method in the recipe is not available. Accepted methods ",
               "are 'basic', 'large_scale', '4nn'."))
        }

        # TO DO: add the possibility to have the element 'pred' in 'data'
        if (lr_method == "large_scale") {
          if (is.null(data$pred$data)) {
            stop("Please provide the large scale predictors in the element 'data$pred$data'.")
          }
        } else {
          data$pred$data <- NULL
        }

        hcst_d <- CST_Intlr(data$hcst, data$obs, 
                            lr_method = lr_method, 
                            target_grid = target_grid, 
                            points = NULL, 
                            int_method = int_methods,
                            method_point_interp = NULL, 
                            predictors = data$pred$data, 
                            lat_dim = "latitude", 
                            lon_dim = "longitude",
                            sdate_dim = "syear", 
                            time_dim = "time", 
                            large_scale_predictor_dimname = 'vars',
                            loocv = loocv, 
                            region = NULL, 
                            ncores = ncores)

        hcst_downscal[[ lr_method ]] <- hcst_d
      }   
 
      DOWNSCAL_MSG <- "##### DOWNSCALING COMPLETE #####"
    } else if (type == "analogs") {
      
      if (is.null(nanalogs)) {
        warning("The number of analogs for searching has not been provided in the ",
                "recipe. Setting it to 3.")
        nanalogs <- 3
      }
      
     
      hcst_downscal <- CST_Analogs(data$hcst, data$obs, 
                                   grid_exp = data$hcst$source_files[1], 
                                   nanalogs = nanalogs, 
                                   fun_analog = NULL, 
                                   lat_dim = "latitude", 
                                   lon_dim = "longitude",
                                   sdate_dim = "syear", 
                                   time_dim = "time", 
                                   region = NULL, 
                                   return_indices = FALSE,
                                   loocv_window = loocv, 
                                   ncores = ncores)

      DOWNSCAL_MSG <- "##### DOWNSCALING COMPLETE #####"
    } else if (type == "logres") {
     
      if (is.null(int_methods) | (length(strsplit(int_methods, ", | |,")[[1]]) != 1)) {
        stop("Please provide one (and only one) interpolation method in the recipe.")
      }

      if (is.null(log_reg_methods)) {
        stop("Please provide one logistic regression method in the recipe. Accepted ", 
              "methods are 'ens_mean', 'ens_mean_sd', 'sorted_members'.")
      }

      if (is.null(target_grid)) {
        stop("Please provide the target grid in the recipe.")
      }
 
      hcst_downscal <- list()
      for (log_reg_method in strsplit(log_reg_methods, ", | |,")[[1]]) { 

        if (!(log_reg_method %in% LOG_REG_METHODS)) {
          stop(paste0(log_reg_method, " method in the recipe is not available. Accepted methods ",
               "are 'ens_mean', 'ens_mean_sd', 'sorted_members'."))
        }

        hcst_d <- CST_LogisticReg(data$hcst, data$obs, 
                                  target_grid = target_grid, 
                                  int_method = int_methods, 
                                  log_reg_method = log_reg_method,
                                  probs_cat = c(1/3,2/3), 
                                  return_most_likely_cat = FALSE, 
                                  points = NULL,
                                  method_point_interp = NULL, 
                                  lat_dim = "latitude", 
                                  lon_dim = "longitude", 
                                  sdate_dim = "syear",
                                  member_dim = "ensemble", 
                                  region = NULL, 
                                  loocv = loocv, 
                                  ncores = ncores)

        hcst_downscal[[ log_reg_method ]] <- hcst_d
      }

      DOWNSCAL_MSG <- "##### DOWNSCALING COMPLETE #####"
    }
  }  
  print(DOWNSCAL_MSG)
  return(list(hcst = hcst_downscal, fcst = NULL))
}   
 
